# --------------------------------------------------------------------------------------------
# Copyright (c) iEXBase. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from trx_utils import (
    is_string,
    remove_0x_prefix,
    is_hex,
    is_integer
)


def is_hex_encoded_block_hash(value):
    if not is_string(value):
        return False
    return len(remove_0x_prefix(value)) == 64 and is_hex(value)


def is_hex_encoded_block_number(value):
    if not is_string(value):
        return False
    elif is_hex_encoded_block_hash(value):
        return False
    try:
        value_as_int = int(value, 16)
    except ValueError:
        return False
    return 0 <= value_as_int < 2**256


def select_method_for_block(value, if_hash, if_number):
    if isinstance(value, bytes):
        return if_hash
    elif is_hex_encoded_block_hash(value):
        return if_hash
    elif is_integer(value) and (0 <= value < 2**256):
        return if_number
    elif is_hex_encoded_block_number(value):
        return if_number
    else:
        raise ValueError(
            "Value did not match any of the recognized block identifiers: {0}".format(value)
        )



# --------------------------------------------------------------------
# Copyright (c) iEXBase. All rights reserved.
# Licensed under the MIT License.
# See License.txt in the project root for license information.
# --------------------------------------------------------------------

import copy

from eth_abi import decode_abi
from eth_utils import (
    function_abi_to_4byte_selector,
    to_hex
)
from hexbytes import HexBytes
from trx_utils import (
    encode_hex,
    is_text,
    deprecated_for,
    combomethod
)

from tronapi.common.abi import (
    filter_by_type,
    merge_args_and_kwargs,
    abi_to_signature,
    fallback_func_abi_exists,
    check_if_arguments_can_be_encoded,
    map_abi_data
)

from tronapi.common.contracts import (
    find_matching_fn_abi,
    encode_abi,
    get_function_info,
    FallbackFn
)
from tronapi.common.datatypes import PropertyCheckingFactory
from tronapi.common.encoding import to_4byte_hex
from tronapi.common.normalizers import (
    normalize_abi,
    normalize_bytecode,
    BASE_RETURN_NORMALIZERS
)
from tronapi.exceptions import (
    NoABIFunctionsFound,
    MismatchedABI,
    FallbackNotFound
)


class NonExistentFallbackFunction:
    @staticmethod
    def _raise_exception():
        raise FallbackNotFound("No fallback function was found in the contract ABI.")

    def __getattr__(self, attr):
        return NonExistentFallbackFunction._raise_exception


class ContractFunction:
    """Base class for contract functions"""
    address = None
    function_identifier = None
    tron = None
    contract_abi = None
    abi = None
    transaction = None
    arguments = None

    def __init__(self, abi=None):
        self.abi = abi
        self.fn_name = type(self).__name__

    def __call__(self, *args, **kwargs):
        clone = copy.copy(self)
        if args is None:
            clone.args = tuple()
        else:
            clone.args = args

        if kwargs is None:
            clone.kwargs = {}
        else:
            clone.kwargs = kwargs
        clone._set_function_info()
        return clone

    def _set_function_info(self):
        if not self.abi:
            self.abi = find_matching_fn_abi(
                self.contract_abi,
                self.function_identifier,
                self.args,
                self.kwargs
            )
        if self.function_identifier is FallbackFn:
            self.selector = encode_hex(b'')
        elif is_text(self.function_identifier):
            self.selector = encode_hex(function_abi_to_4byte_selector(self.abi))
        else:
            raise TypeError("Unsupported function identifier")

        self.arguments = merge_args_and_kwargs(self.abi, self.args, self.kwargs)

    @classmethod
    def factory(cls, class_name, **kwargs):
        return PropertyCheckingFactory(class_name, (cls,), kwargs)(kwargs.get('abi'))

    def __repr__(self):
        if self.abi:
            _repr = '<Function %s' % abi_to_signature(self.abi)
            if self.arguments is not None:
                _repr += ' bound to %r' % (self.arguments,)
            return _repr + '>'
        return '<Function %s>' % self.fn_name


class ContractFunctions:
    """Class containing contract function objects
    """

    def __init__(self, abi, tron, address=None):
        if abi:
            self.abi = abi
            self._functions = filter_by_type('function', self.abi)
            for func in self._functions:
                setattr(
                    self,
                    func['name'],
                    ContractFunction.factory(
                        func['name'],
                        tron=tron,
                        contract_abi=self.abi,
                        address=address,
                        function_identifier=func['name']))

    def __iter__(self):
        if not hasattr(self, '_functions') or not self._functions:
            return

        for func in self._functions:
            yield func['name']

    def __getattr__(self, function_name):
        if '_functions' not in self.__dict__:
            raise NoABIFunctionsFound(
                "The abi for this contract contains no function definitions. ",
                "Are you sure you provided the correct contract abi?"
            )
        elif function_name not in self.__dict__['_functions']:
            raise MismatchedABI(
                "The function '{}' was not found in this contract's abi. ".format(function_name),
                "Are you sure you provided the correct contract abi?"
            )
        else:
            return super().__getattribute__(function_name)

    def __getitem__(self, function_name):
        return getattr(self, function_name)


class Contract:
    # set during class construction
    tron = None

    # instance level properties
    address = None

    # class properties (overridable at instance level)
    abi = None

    bytecode = None
    bytecode_runtime = None

    functions = None
    events = None

    def __init__(self, address=None):
        """Create a new smart contract proxy object.
        :param address: Contract address as 0x hex string
        """
        if self.tron is None:
            raise AttributeError(
                'The `Contract` class has not been initialized.  Please use the '
                '`tron.contract` interface to create your contract class.'
            )

        if address:
            self.address = self.tron.address.to_hex(address)

        if not self.address:
            raise TypeError("The address argument is required to instantiate a contract.")

        self.functions = ContractFunctions(self.abi, self.tron, self.address)
        self.fallback = Contract.get_fallback_function(self.abi, self.tron, self.address)

    @classmethod
    def factory(cls, tron, class_name=None, **kwargs):

        kwargs['tron'] = tron
        normalizers = {
            'abi': normalize_abi,
            'bytecode': normalize_bytecode,
            'bytecode_runtime': normalize_bytecode,
        }

        contract = PropertyCheckingFactory(
            class_name or cls.__name__,
            (cls,),
            kwargs,
            normalizers=normalizers
        )

        setattr(contract, 'functions', ContractFunctions(contract.abi, contract.tron))
        setattr(contract, 'fallback', Contract.get_fallback_function(contract.abi, contract.tron))

        return contract

    @classmethod
    @deprecated_for("contract.constructor.transact")
    def deploy(cls, **kwargs):
        """Deploy Contract

        Deploys a contract.
        Returns TransactionExtention, which contains an unsigned transaction.

        Example:
        .. code-block:: python
            >>> MyContract.deploy(
                fee_limit=10**9,
                call_value=0,
                consume_user_resource_percent=10
            )

        Args:
            **kwargs: Transaction parameters for the deployment
            transaction as a dict

        """
        return cls.tron.transaction_builder.create_smart_contract(
            **kwargs,
            abi=cls.abi,
            bytecode=to_hex(cls.bytecode)
        )

    @classmethod
    def constructor(cls):
        if cls.bytecode is None:
            raise ValueError(
                "Cannot call constructor on a contract that does not have 'bytecode' associated "
                "with it"
            )
        return ContractConstructor(cls.tron,
                                   cls.abi,
                                   cls.bytecode)

    @combomethod
    def encodeABI(cls, fn_name, args=None, kwargs=None, data=None):
        """Encodes the arguments using the Tron ABI for the contract function
        that matches the given name and arguments..
        """
        fn_abi, fn_selector, fn_arguments = get_function_info(
            fn_name, contract_abi=cls.abi, args=args, kwargs=kwargs,
        )

        if data is None:
            data = fn_selector

        return encode_abi(cls.tron, fn_abi, fn_arguments, data)

    @staticmethod
    def get_fallback_function(abi, tron, address=None):
        if abi and fallback_func_abi_exists(abi):
            return ContractFunction.factory(
                'fallback',
                tron=tron,
                contract_abi=abi,
                address=address,
                function_identifier=FallbackFn)()

        return NonExistentFallbackFunction()

    @combomethod
    def all_functions(self):
        return find_functions_by_identifier(
            self.abi, self.tron, self.address, lambda _: True
        )

    @combomethod
    def get_function_by_signature(self, signature):
        if ' ' in signature:
            raise ValueError(
                'Function signature should not contain any spaces. '
                'Found spaces in input: %s' % signature
            )

        def callable_check(fn_abi):
            return abi_to_signature(fn_abi) == signature

        fns = find_functions_by_identifier(self.abi, self.tron, self.address, callable_check)
        return get_function_by_identifier(fns, 'signature')

    @combomethod
    def find_functions_by_name(self, fn_name):
        def callable_check(fn_abi):
            return fn_abi['name'] == fn_name

        return find_functions_by_identifier(
            self.abi, self.tron, self.address, callable_check
        )

    @combomethod
    def get_function_by_name(self, fn_name):
        fns = self.find_functions_by_name(fn_name)
        return get_function_by_identifier(fns, 'name')

    @combomethod
    def get_function_by_selector(self, selector):
        def callable_check(fn_abi):
            return encode_hex(function_abi_to_4byte_selector(fn_abi)) == to_4byte_hex(selector)

        fns = find_functions_by_identifier(self.abi, self.tron, self.address, callable_check)
        return get_function_by_identifier(fns, 'selector')

    @combomethod
    def decode_function_input(self, data):
        data = HexBytes(data)
        selector, params = data[:4], data[4:]
        func = self.get_function_by_selector(selector)
        names = [x['name'] for x in func.abi['inputs']]
        types = [x['type'] for x in func.abi['inputs']]
        decoded = decode_abi(types, params)
        normalized = map_abi_data(BASE_RETURN_NORMALIZERS, types, decoded)
        return func, dict(zip(names, normalized))

    @combomethod
    def find_functions_by_args(self, *args):
        def callable_check(fn_abi):
            return check_if_arguments_can_be_encoded(fn_abi, args=args, kwargs={})

        return find_functions_by_identifier(
            self.abi, self.tron, self.address, callable_check
        )

    @combomethod
    def get_function_by_args(self, *args):
        fns = self.find_functions_by_args(*args)
        return get_function_by_identifier(fns, 'args')


class ContractConstructor:
    """
    Class for contract constructor API.
    """

    def __init__(self, tron, abi, bytecode):
        self.tron = tron
        self.abi = abi
        self.bytecode = bytecode

    @staticmethod
    def check_forbidden_keys_in_transaction(transaction, forbidden_keys=None):
        keys_found = set(transaction.keys()) & set(forbidden_keys)
        if keys_found:
            raise ValueError("Cannot set {} in transaction".format(', '.join(keys_found)))

    @combomethod
    def transact(self, **kwargs):
        """Deploy Contract

        Deploys a contract.
        Returns TransactionExtention, which contains an unsigned transaction.

        Args:
            **kwargs: Additional options to send
        """

        return self.tron.transaction_builder.create_smart_contract(
            **kwargs,
            abi=self.abi,
            bytecode=to_hex(self.bytecode)
        )


def find_functions_by_identifier(contract_abi, tron, address, callable_check):
    fns_abi = filter_by_type('function', contract_abi)
    return [
        ContractFunction.factory(
            fn_abi['name'],
            tron=tron,
            contract_abi=contract_abi,
            address=address,
            function_identifier=fn_abi['name'],
            abi=fn_abi
        )
        for fn_abi in fns_abi
        if callable_check(fn_abi)
    ]


def get_function_by_identifier(fns, identifier):
    if len(fns) > 1:
        raise ValueError(
            'Found multiple functions with matching {0}. '
            'Found: {1!r}'.format(identifier, fns)
        )
    elif len(fns) == 0:
        raise ValueError(
            'Could not find any function with matching {0}'.format(identifier)
        )
    return fns[0]



# --------------------------------------------------------------------
# Copyright (c) iEXBase. All rights reserved.
# Licensed under the MIT License.
# See License.txt in the project root for license information.
# --------------------------------------------------------------------

import sys

import pkg_resources

from eth_account import Account  # noqa: E402
from tronapi.providers.http import HttpProvider  # noqa: E402
from tronapi.main import Tron  # noqa: E402

if sys.version_info < (3, 5):
    raise EnvironmentError("Python 3.5 or above is required")


__version__ = pkg_resources.get_distribution("tronapi-JCO").version

__all__ = [
    "__version__",
    "HttpProvider",
    "Account",
    "Tron",
]



# --------------------------------------------------------------------
# Copyright (c) iEXBase. All rights reserved.
# Licensed under the MIT License.
# See License.txt in the project root for license information.
# --------------------------------------------------------------------

import codecs
from binascii import unhexlify

import base58
import ecdsa
from eth_keys import KeyAPI
from eth_account import Account as ETHAccount
from trx_utils import is_hex, is_bytes

from tronapi.common.datastructures import AttributeDict


class Account:
    @staticmethod
    def create():
        generate_key = ecdsa.SigningKey.generate(curve=ecdsa.SECP256k1)
        return PrivateKey(generate_key.to_string().hex())

    @staticmethod
    def sign_hash(message_hash, private_key):
        if not is_hex(message_hash):
            raise ValueError('Invalid message_hash provided')

        return ETHAccount.signHash(message_hash, private_key)

    @staticmethod
    def recover_hash(message_hash, signature):
        if not is_hex(message_hash):
            raise ValueError('Invalid message_hash provided')

        return ETHAccount.recoverHash(message_hash, signature=signature)


class Address(object):
    @staticmethod
    def from_hex(address):
        """Helper function that will convert a generic value from hex"""
        if not is_hex(address):
            return address

        return base58.b58encode_check(bytes.fromhex(address))

    @staticmethod
    def to_hex(address):
        """Helper function that will convert a generic value to hex"""
        if is_hex(address):
            return address.lower().replace('0x', '41', 2)

        return base58.b58decode_check(address).hex().upper()

    @staticmethod
    def from_private_key(private_key):
        return PrivateKey(private_key).address


class PrivateKey(object):
    def __init__(self, private_key):
        """Work with private key.
        Getting: PublicKey, PublicToAddress

        Example:::
            PrivateKey("4d1bc37b069b9f2e975c37770b7c87185dc3a10454e3ea024ce1fce8f3eb78bf")
        """
        _private = unhexlify(bytes(private_key, encoding='utf8'))
        self._key = KeyAPI.PrivateKey(_private)
        _length = len(self._key)

        # Key length must not exceed 64 length
        if _length < 64:
            raise ValueError('Key length must not exceed 64 length')

    @property
    def private_key(self):
        _raw_key = self._key.to_bytes()
        return codecs.decode(codecs.encode(_raw_key, 'hex'), 'ascii')

    @property
    def public_key(self) -> str:
        public_key = self._key.public_key
        return '04' + str(public_key)[2:]

    @property
    def address(self):
        public_key = self._key.public_key
        address = '41' + public_key.to_address()[2:]
        to_base58 = base58.b58encode_check(bytes.fromhex(address))

        # If bytecode then convert to string
        if is_bytes(to_base58):
            to_base58 = to_base58.decode()

        return AttributeDict({
            'hex': address,
            'base58': to_base58
        })

    def __str__(self):
        return self.private_key

    def __bytes__(self):
        return self._key.to_bytes()



# --------------------------------------------------------------------
# Copyright (c) iEXBase. All rights reserved.
# Licensed under the MIT License.
# See License.txt in the project root for license information.
# --------------------------------------------------------------------

"""
    tronapi.trx
    ===============

    Work with basic methods

    :copyright: © 2018 by the iEXBase.
    :license: MIT License
"""

import math
from typing import Any

from trx_utils import is_integer, is_hex
from trx_utils.types import is_object, is_string, is_list, is_boolean

from tronapi.common.transactions import wait_for_transaction_id
from tronapi.contract import Contract
from tronapi.exceptions import InvalidTronError, TronError, TimeExhausted
from tronapi.module import Module
from tronapi.common.blocks import select_method_for_block
from tronapi.common.toolz import assoc
from tronapi.common.account import Account

TRX_MESSAGE_HEADER = "\x19TRON Signed Message:\n"
ETH_MESSAGE_HEADER = "\x19Ethereum Signed Message:\n"


class Trx(Module):
    default_contract_factory = Contract

    def get_current_block(self):
        """Query the latest block"""
        return self.tron.manager.request(url="/wallet/getnowblock")

    def get_confirmed_current_block(self):
        """Query the confirmed latest block"""
        return self.tron.manager.request("/walletsolidity/getnowblock")

    def get_block(self, block: Any = None):
        """Get block details using HashString or blockNumber

        Args:
            block (Any): ID or height for the block

        """

        # If the block identifier is not specified,
        # we take the default
        if block is None:
            block = self.tron.default_block

        if block == "latest":
            return self.get_current_block()
        elif block == "earliest":
            return self.get_block(0)

        method = select_method_for_block(
            block,
            if_hash={"url": "/wallet/getblockbyid", "field": "value"},
            if_number={"url": "/wallet/getblockbynum", "field": "num"},
        )

        result = self.tron.manager.request(method["url"], {method["field"]: block})

        if result:
            return result
        raise ValueError("The call to {method['url']} did not return a value.")

    def get_transaction_count_by_blocknum(self, num: int):
        """Query transaction's count on a specified block by height

        Args:
            num (int): block number
        """
        if not is_integer(num) or num < 0:
            raise ValueError("Invalid num provided")

        return self.tron.manager.request(
            "/wallet/gettransactioncountbyblocknum", {"num": num}
        )

    def get_block_transaction_count(self, block: Any):
        """Total number of transactions in a block

        Args:
            block (Any): Number or Hash Block

        """
        transaction = self.get_block(block)
        if "transactions" not in transaction:
            raise TronError('Parameter "transactions" not found')

        return len(transaction)

    def get_transaction_from_block(self, block: Any, index: int = 0):
        """Get transaction details from Block

        Args:
            block (Any): Number or Hash Block
            index (int) Position

        """
        if not is_integer(index) or index < 0:
            raise InvalidTronError("Invalid transaction index provided")

        transactions = self.get_block(block).get("transactions")
        if not transactions or len(transactions) < index:
            raise TronError("Transaction not found in block")

        return transactions[index]

    def wait_for_transaction_id(
        self, transaction_hash: str, timeout=120, poll_latency=0.2
    ):
        """
        Waits for the transaction specified by transaction_hash
        to be included in a block, then returns its transaction receipt.

        Optionally, specify a timeout in seconds.
        If timeout elapses before the transaction is added to a block,
        then wait_for_transaction_id() raises a Timeout exception.


        Args:
            transaction_hash (str): Transaction Hash
            timeout (int): TimeOut
            poll_latency (any):  between subsequent requests

        """
        try:
            if poll_latency > timeout:
                poll_latency = timeout

            return wait_for_transaction_id(
                self.tron, transaction_hash, timeout, poll_latency
            )
        except TimeoutError:
            raise TimeExhausted(
                "Transaction {} is not in the chain, after {} seconds".format(
                    transaction_hash,
                    timeout,
                )
            )

    def get_transaction(self, transaction_id: str, is_confirm: bool = False):
        """Query transaction based on id

        Args:
            transaction_id (str): transaction id
            is_confirm (bool):
        """

        method = "walletsolidity" if is_confirm else "wallet"
        response = self.tron.manager.request(
            "/{}/gettransactionbyid".format(method), {"value": transaction_id}
        )

        if "txID" not in response:
            raise ValueError("Transaction not found")

        return response

    def get_account_by_id(self, account_id: str, options: object):
        return self.get_account_info_by_id(account_id, options)

    def get_account_info_by_id(self, account_id: str, options: object):
        if account_id.startswith("0x"):
            account_id = id[2:]

        if "confirmed" in options:
            return self.tron.manager.request(
                "/walletsolidity/getaccountbyid",
                {"account_id": self.tron.toHex(text=account_id)},
            )

        return self.tron.manager.request(
            "/wallet/getaccountbyid", {"account_id": self.tron.toHex(text=account_id)}
        )

    def get_unconfirmed_account_by_id(self, account_id: str):
        return self.get_account_info_by_id(account_id, {"confirmed": True})

    def get_account_resource(self, address=None):
        """Query the resource information of the account

        Args:
            address (str): Address

        Results:
            Resource information of the account

        """

        if address is None:
            address = self.tron.default_address.hex

        if not self.tron.isAddress(address):
            raise InvalidTronError("Invalid address provided")

        return self.tron.manager.request(
            "/wallet/getaccountresource", {"address": self.tron.address.to_hex(address)}
        )

    def get_account(self, address=None):
        """Query information about an account

        Args:
            address (str): Address

        """

        if address is None:
            address = self.tron.default_address.hex

        if not self.tron.isAddress(address):
            raise InvalidTronError("Invalid address provided")

        return self.tron.manager.request(
            "/walletsolidity/getaccount", {"address": self.tron.address.to_hex(address)}
        )

    def get_balance(self, address=None, is_float=False):
        """Getting a balance

        Args:
            address (str): Address
            is_float (bool): Convert to float format

        """
        response = self.get_account(address)
        if "balance" not in response:
            return 0

        if is_float:
            return self.tron.fromSun(response["balance"])

        return response["balance"]

    def get_transactions_related(self, address, direction="all", limit=30, offset=0):
        """Getting data in the "from", "to" and "all" directions

        Args:
            address (str): Address
            direction (str): Type direction
            address (str): address
            limit (int): number of transactions expected to be returned
            offset (int): index of the starting transaction

        """

        if direction not in ["from", "to", "all"]:
            raise InvalidTronError(
                'Invalid direction provided: Expected "to", "from" or "all"'
            )

        if direction == "all":
            _from = self.get_transactions_related(address, "from", limit, offset)
            _to = self.get_transactions_related(address, "to", limit, offset)

            filter_from = [{**i, "direction": "from"} for i in _from]
            filter_to = [{**i, "direction": "to"} for i in _to]

            callback = filter_from
            callback.extend(filter_to)
            return callback

        if address is None:
            address = self.tron.default_address.hex

        if not self.tron.isAddress(address):
            raise InvalidTronError("Invalid address provided")

        if not isinstance(limit, int) or limit < 0 or (offset and limit < 1):
            raise InvalidTronError("Invalid limit provided")

        if not isinstance(offset, int) or offset < 0:
            raise InvalidTronError("Invalid offset provided")

        path = "/walletextension/gettransactions{0}this".format(direction)
        response = self.tron.manager.request(
            path,
            {
                "account": {"address": self.tron.address.to_hex(address)},
                "limit": limit,
                "offset": offset,
            },
            "get",
        )

        if "transaction" in response:
            return response["transaction"]
        return response

    def get_transactions_to_address(self, address=None, limit=30, offset=0):
        """Query the list of transactions received by an address

        Args:
            address (str): address
            limit (int): number of transactions expected to be returned
            offset (int): index of the starting transaction

        Returns:
            Transactions list

        """
        return self.get_transactions_related(address, "to", limit, offset)

    def get_transactions_from_address(self, address=None, limit=30, offset=0):
        """Query the list of transactions sent by an address

        Args:
            address (str): address
            limit (int): number of transactions expected to be returned
            offset (int): index of the starting transaction

        Returns:
            Transactions list

        """
        return self.get_transactions_related(address, "from", limit, offset)

    def get_transaction_info(self, tx_id):
        """Query transaction fee based on id

        Args:
            tx_id (str): Transaction Id

        Returns:
            Transaction fee，block height and block creation time

        """
        response = self.tron.manager.request(
            "/walletsolidity/gettransactioninfobyid", {"value": tx_id}
        )

        return response

    def get_band_width(self, address=None):
        """Query bandwidth information.

        Args:
            address (str): address

        Returns:
            Bandwidth information for the account.
            If a field doesn't appear, then the corresponding value is 0.
            {
                "freeNetUsed": 557,
                "freeNetLimit": 5000,
                "NetUsed": 353,
                "NetLimit": 5239157853,
                "TotalNetLimit": 43200000000,
                "TotalNetWeight": 41228
            }

        """

        if address is None:
            address = self.tron.default_address.hex

        if not self.tron.isAddress(address):
            raise InvalidTronError("Invalid address provided")

        response = self.tron.manager.request(
            "/wallet/getaccountnet", {"address": self.tron.address.to_hex(address)}
        )

        free_net_limit = (
            0 if "freeNetLimit" not in response else response["freeNetLimit"]
        )
        free_net_used = 0 if "freeNetUsed" not in response else response["freeNetUsed"]
        net_limit = 0 if "NetLimit" not in response else response["NetLimit"]
        net_used = 0 if "NetUsed" not in response else response["NetUsed"]

        return (free_net_limit - free_net_used) + (net_limit - net_used)

    def get_transaction_count(self):
        """Count all transactions on the network
        Note: Possible delays

        Returns:
            Total number of transactions.

        """
        response = self.tron.manager.request("/wallet/totaltransaction")
        return response.get("num")

    def send(self, to, amount, options=None):
        """Send funds to the Tron account (option 2)"""
        return self.send_transaction(to, amount, options)

    def send_trx(self, to, amount, options=None):
        """Send funds to the Tron account (option 3)"""
        return self.send_transaction(to, amount, options)

    def send_transaction(self, to, amount, options=None):
        """Send an asset to another account.
        Will create and broadcast the transaction if a private key is provided.

        Args:
            to (str): Address to send TRX to.
            amount (float): Amount of TRX to send.
            options (Any, optional): Options

        """

        if options is None:
            options = {}

        if "from" not in options:
            options = assoc(options, "from", self.tron.default_address.hex)

        tx = self.tron.transaction_builder.send_transaction(to, amount, options["from"])
        # If a comment is attached to the transaction,
        # in this case adding to the object
        if "message" in options:
            tx["raw_data"]["data"] = self.tron.toHex(text=str(options["message"]))

        sign = self.sign(tx)
        result = self.broadcast(sign)

        return result

    def send_token(self, to, amount, token_id=None, account=None):
        """Transfer Token

        Args:
            to (str): is the recipient address
            amount (float): is the amount of token to transfer
            token_id (str): Token Name(NOT SYMBOL)
            account: (str): is the address of the withdrawal account

        Returns:
            Token transfer Transaction raw data

        """
        if account is None:
            account = self.tron.default_address.hex

        tx = self.tron.transaction_builder.send_token(to, amount, token_id, account)
        sign = self.sign(tx)
        result = self.broadcast(sign)

        return result

    def freeze_balance(self, amount=0, resource="BANDWIDTH", account=None):
        """
        Freezes an amount of TRX.
        Will give bandwidth OR Energy and TRON Power(voting rights)
        to the owner of the frozen tokens.

        Args:
            amount (int): number of frozen trx
            duration (int): duration in days to be frozen
            resource (str): type of resource, must be either "ENERGY" or "BANDWIDTH"
            account (str): address that is freezing trx account

        """

        if account is None:
            account = self.tron.default_address.hex

        transaction = self.tron.transaction_builder.freeze_balance(
            amount, resource, account
        )
        sign = self.sign(transaction)
        response = self.broadcast(sign)

        return response

    def unfreeze_balance(self, amount=0, resource="BANDWIDTH", account=None):
        """
        Unfreeze TRX that has passed the minimum freeze duration.
        Unfreezing will remove bandwidth and TRON Power.

        Args:
            resource (str): type of resource, must be either "ENERGY" or "BANDWIDTH"
            account (str): address that is freezing trx account

        """

        if account is None:
            account = self.tron.default_address.hex

        transaction = self.tron.transaction_builder.unfreeze_balance(
            amount, resource, account
        )
        sign = self.sign(transaction)
        response = self.broadcast(sign)

        return response

    def withdraw_expire_unfreeze(self, account=None):
        """
        Withdraw unfrozen balance in Stake2.0.
        the user can call this API to get back their funds.
        After executing /wallet/unfreezebalancev2 transaction and waiting N days, N is a network parameter

        Args:
            account (str): address that owns the staked trx

        """

        if account is None:
            account = self.tron.default_address.hex

        transaction = self.tron.transaction_builder.withdraw_expire_unfreeze(account)
        sign = self.sign(transaction)
        response = self.broadcast(sign)

        return response

    def get_can_withdraw_unfreeze_amount(self, account=None):
        """
        Get amount of available unfreeze amount to withdraw in Stake2.0.
        the user can call this API to know how much funds they can get back.

        Args:
            account (str): address that owns the staked trx

        """

        if account is None:
            account = self.tron.default_address.hex

        transaction = self.tron.transaction_builder.get_can_withdraw_unfreeze_amount(
            account
        )
        sign = self.sign(transaction)
        response = self.broadcast(sign)

        return response

    def delegate_resource(
        self, amount=0, resource="BANDWIDTH", receiver=None, lock=False, account=None
    ):
        """
        Delegate resource to another account.
        Will give bandwidth OR Energy and TRON Power(voting rights)
        to the receiver of the frozen tokens.

        Args:
            amount (int): number of frozen trx
            resource (str): type of resource, must be either "ENERGY" or "BANDWIDTH"
            receiver (str): address that will receive the resource
            lock (int): whether to lock the resource
            account (str): address that owns the resource

        """

        if account is None:
            account = self.tron.default_address.hex

        transaction = self.tron.transaction_builder.delegate_resource(
            amount, resource, receiver, lock, account
        )
        sign = self.sign(transaction)
        response = self.broadcast(sign)

        return response

    def get_can_delegated_max_size(
        self, owner_address=None, resource="BANDWIDTH", visible=True
    ):
        """
        Get maximum amount of available resource to delegate.

        Args:
            owner_address (int): number of frozen trx
            resource (str): type of resource, must be either "ENERGY" or "BANDWIDTH"
            visible (str): Whether the address is in base58 format

        """

        if owner_address is None:
            owner_address = self.tron.default_address.hex

        if resource not in (
            "BANDWIDTH",
            "ENERGY",
        ):
            raise InvalidTronError(
                'Invalid resource provided: Expected "BANDWIDTH" or "ENERGY"'
            )

        if not self.tron.isAddress(owner_address):
            raise InvalidTronError("Invalid address provided")

        if not is_boolean(visible):
            raise InvalidTronError("Invalid visible provided")

        response =  self.tron.manager.request(
            "/wallet/getcandelegatedmaxsize",
            {
                "owner_address": self.tron.address.to_hex(owner_address),
                "type": 1 if resource == "ENERGY" else 0 if resource == "BANDWIDTH" else None,
                "visible": visible,
            },
        )

        return {"max_size": self.tron.fromSun(response["max_size"])}

    def undelegate_resource(
        self, amount=0, resource="BANDWIDTH", receiver=None, account=None
    ):
        """
        Undelegate resource to another account.
        Will get back bandwidth OR Energy and TRON Power(voting rights)
        from the receiver of the frozen tokens.

        Args:
            amount (int): number of frozen trx
            resource (str): type of resource, must be either "ENERGY" or "BANDWIDTH"
            receiver (str): address that received the resource
            account (str): address that owns the resource

        """

        if account is None:
            account = self.tron.default_address.hex

        transaction = self.tron.transaction_builder.undelegate_resource(
            amount, resource, receiver, account
        )
        sign = self.sign(transaction)
        response = self.broadcast(sign)

        return response

    def online_sign(self, transaction: dict):
        """Online transaction signature
        Sign the transaction, the api has the risk of leaking the private key,
        please make sure to call the api in a secure environment

        Warnings:
            Do not use this in any web / user-facing applications.
            This will expose the private key.

        Args:
            transaction (dict): transaction details

        """

        if "signature" in transaction:
            raise TronError("Transaction is already signed")

        address = self.tron.address.from_private_key(self.tron.private_key).hex.lower()
        owner_address = transaction["raw_data"]["contract"][0]["parameter"]["value"][
            "owner_address"
        ]

        if address != owner_address:
            raise ValueError("Private key does not match address in transaction")

        return self.tron.manager.request(
            "/wallet/gettransactionsign",
            {"transaction": transaction, "privateKey": self.tron.private_key},
        )

    def sign(self, transaction: Any, use_tron: bool = True, multisig: bool = False):
        """Safe method for signing your transaction

        Warnings:
            method: online_sign() - Use only in extreme cases.

        Args:
            transaction (Any): transaction details
            use_tron (bool): is Tron header
            multisig (bool): multi sign

        """

        if is_string(transaction):
            if not is_hex(transaction):
                raise TronError("Expected hex message input")

            # Determine which header to attach to the message
            # before encrypting or decrypting
            header = TRX_MESSAGE_HEADER if use_tron else ETH_MESSAGE_HEADER
            header += str(len(transaction))

            message_hash = self.tron.keccak(text=header + transaction)

            signed_message = Account.sign_hash(
                self.tron.toHex(message_hash), self.tron.private_key
            )
            return signed_message

        if not multisig and "signature" in transaction:
            raise TronError("Transaction is already signed")

        try:
            if not multisig:
                address = self.tron.address.from_private_key(
                    self.tron.private_key
                ).hex.lower()
                owner_address = transaction["raw_data"]["contract"][0]["parameter"][
                    "value"
                ]["owner_address"]

                if address != owner_address:
                    raise ValueError(
                        "Private key does not match address in transaction"
                    )

            # This option deals with signing of transactions, and writing to the array
            signed_tx = Account.sign_hash(transaction["txID"], self.tron.private_key)
            signature = signed_tx["signature"].hex()[2:]

            # support multi sign
            if "signature" in transaction and is_list(transaction["signature"]):
                if not transaction["signature"].index(signature):
                    transaction["signature"].append(signature)
            else:
                transaction["signature"] = [signature]

            return transaction
        except ValueError as err:
            raise InvalidTronError(err)

    def broadcast(self, signed_transaction):
        """Broadcast the signed transaction

        Args:
            signed_transaction (object): signed transaction contract data

        """
        if not is_object(signed_transaction):
            raise InvalidTronError("Invalid transaction provided")

        if "signature" not in signed_transaction:
            raise TronError("Transaction is not signed")

        response = self.tron.manager.request(
            "/wallet/broadcasttransaction", signed_transaction
        )

        if "result" in response:
            response.update({"transaction": signed_transaction})
        return response

    def sign_and_broadcast(self, transaction: Any):
        """Sign and send to the network

        Args:
            transaction (Any): transaction details
        """
        if not is_object(transaction):
            raise TronError("Invalid transaction provided")

        signed_tx = self.sign(transaction)
        return self.broadcast(signed_tx)

    def verify_message(
        self, message, signed_message=None, address=None, use_tron: bool = True
    ):
        """Get the address of the account that signed the message with the given hash.
        You must specify exactly one of: vrs or signature

        Args:
            message (str): The message in the format "hex"
            signed_message (AttributeDict): Signature
            address (str): is Address
            use_tron (bool): is Tron header

        """
        if address is None:
            address = self.tron.default_address.base58

        if not is_hex(message):
            raise TronError("Expected hex message input")

        # Determine which header to attach to the message
        # before encrypting or decrypting
        header = TRX_MESSAGE_HEADER if use_tron else ETH_MESSAGE_HEADER
        header += str(len(message))

        message_hash = self.tron.keccak(text=header + message)
        recovered = Account.recover_hash(
            self.tron.toHex(message_hash), signed_message.signature
        )

        tron_address = "41" + recovered[2:]
        base58address = self.tron.address.from_hex(tron_address).decode()

        if base58address == address:
            return True

        raise ValueError("Signature does not match")

    def update_account(self, account_name, address=None):
        """Modify account name
        Note: Username is allowed to edit only once.

        Args:
            account_name (str): name of the account
            address (str): address

        """
        if address is None:
            address = self.tron.default_address.hex

        transaction = self.tron.transaction_builder.update_account(
            account_name, address
        )
        sign = self.sign(transaction)
        response = self.broadcast(sign)

        return response

    def apply_for_sr(self, url, address):
        """Apply to become a super representative
        Note: Applied to become a super representative. Cost 9999 TRX.

        Args:
            url (str): official website address
            address (str): address

        """

        if address is None:
            address = self.tron.default_address.hex

        transaction = self.tron.transaction_builder.apply_for_sr(url, address)
        sign = self.sign(transaction)
        response = self.broadcast(sign)

        return response

    def list_nodes(self):
        """List the nodes which the api fullnode is connecting on the network"""
        response = self.tron.manager.request("/wallet/listnodes")
        callback = map(
            lambda x: {
                "address": "{}:{}".format(
                    self.tron.toText(x["address"]["host"]), str(x["address"]["port"])
                )
            },
            response["nodes"],
        )

        return list(callback)

    def get_tokens_issued_by_address(self, address):
        """List the tokens issued by an account.

        Args:
            address (str): address

        Returns:
            The token issued by the account.
            An account can issue only one token.

        """

        if not self.tron.isAddress(address):
            raise InvalidTronError("Invalid address provided")

        address = self.tron.address.to_hex(address)

        return self.tron.manager.request(
            "/wallet/getassetissuebyaccount", {"address": address}
        )

    def get_token_from_id(self, token_id: str):
        """Query token by name.

        Args:
            token_id (str): The name of the token

        """
        if not isinstance(token_id, str) or not len(token_id):
            raise InvalidTronError("Invalid token ID provided")

        return self.tron.manager.request(
            "/wallet/getassetissuebyname", {"value": self.tron.toHex(text=token_id)}
        )

    def get_block_range(self, start, end):
        """Query a range of blocks by block height

        Args:
            start (int): starting block height, including this block
            end (int): ending block height, excluding that block

        """
        if not is_integer(start) or start < 0:
            raise InvalidTronError("Invalid start of range provided")

        if not is_integer(end) or end <= start:
            raise InvalidTronError("Invalid end of range provided")

        response = self.tron.manager.request(
            "/wallet/getblockbylimitnext",
            {"startNum": int(start), "endNum": int(end) + 1},
            "post",
        )

        return response.get("block")

    def get_latest_blocks(self, num=1):
        """Query the latest blocks

        Args:
            num (int): the number of blocks to query

        """
        if not is_integer(num) or num <= 0:
            raise InvalidTronError("Invalid limit provided")

        response = self.tron.manager.request(
            "/wallet/getblockbylatestnum", {"num": num}
        )

        return response.get("block")

    def list_super_representatives(self):
        """Query the list of Super Representatives"""
        response = self.tron.manager.request("/wallet/listwitnesses")
        return response.get("witnesses")

    def list_tokens(self, limit=0, offset=0):
        """Query the list of Tokens with pagination

        Args:
            limit (int): index of the starting Token
            offset (int): number of Tokens expected to be returned

        Returns:
            List of Tokens

        """
        if not is_integer(limit) or (limit and offset < 1):
            raise InvalidTronError("Invalid limit provided")

        if not is_integer(offset) or offset < 0:
            raise InvalidTronError("Invalid offset provided")

        if not limit:
            return self.tron.manager.request("/wallet/getassetissuelist").get(
                "assetIssue"
            )

        return self.tron.manager.request(
            "/wallet/getpaginatedassetissuelist",
            {"limit": int(limit), "offset": int(offset)},
        )

    def time_until_next_vote_cycle(self):
        """Get the time of the next Super Representative vote

        Returns:
            Number of milliseconds until the next voting time.

        """
        num = self.tron.manager.request("/wallet/getnextmaintenancetime").get("num")

        if num == -1:
            raise Exception("Failed to get time until next vote cycle")

        return math.floor(num / 1000)

    def get_contract(self, contract_address):
        """Queries a contract's information from the blockchain.

        Args:
            contract_address (str): contract address

        Returns:
            SmartContract object.

        """

        if not self.tron.isAddress(contract_address):
            raise InvalidTronError("Invalid contract address provided")

        return self.tron.manager.request(
            "/wallet/getcontract", {"value": self.tron.address.to_hex(contract_address)}
        )

    def contract(self, address=None, **kwargs):
        """Work with a contract

        Args:
            address (str): TRON Address
            **kwargs (any): details (bytecode, abi)
        """
        factory_class = kwargs.pop(
            "contract_factory_class", self.default_contract_factory
        )
        contract_factory = factory_class.factory(self.tron, **kwargs)

        if address:
            return contract_factory(address)
        return contract_factory

    def validate_address(self, address, _is_hex=False):
        """Validate address

        Args:
            address (str): The address, should be in base58checksum
            _is_hex (bool): hexString or base64 format

        """
        if _is_hex:
            address = self.tron.address.to_hex(address)

        return self.tron.manager.request(
            "/wallet/validateaddress", {"address": address}
        )

    def get_chain_parameters(self):
        """Getting chain parameters"""
        return self.tron.manager.request("/wallet/getchainparameters")

    def get_exchange_by_id(self, exchange_id):
        """Find exchange by id

        Args:
             exchange_id (str): ID Exchange

        """

        if not isinstance(exchange_id, int) or exchange_id < 0:
            raise InvalidTronError("Invalid exchangeID provided")

        return self.tron.manager.request("/wallet/getexchangebyid", {"id": exchange_id})

    def get_list_exchangers(self):
        """Get list exchangers"""
        return self.tron.manager.request("/wallet/listexchanges")

    def get_proposal(self, proposal_id):
        """Query proposal based on id

        Args:
            proposal_id (int): ID

        """
        if not isinstance(proposal_id, int) or proposal_id < 0:
            raise InvalidTronError("Invalid proposalID provided")

        return self.tron.manager.request(
            "/wallet/getproposalbyid", {"id": int(proposal_id)}
        )

    def list_proposals(self):
        """Query all proposals

        Returns:
            Proposal list information

        """
        return self.tron.manager.request("/wallet/listproposals")

    def vote_proposal(self, proposal_id, has_approval, voter_address):
        """Proposal approval

        Args:
            proposal_id (int): proposal id
            has_approval (bool): Approved
            voter_address (str): Approve address

        Returns:
             Approval of the proposed transaction

        """

        if voter_address is None:
            voter_address = self.tron.default_address.hex

        transaction = self.tron.transaction_builder.vote_proposal(
            proposal_id, has_approval, voter_address
        )
        sign = self.sign(transaction)
        response = self.broadcast(sign)

        return response

    def proposal_delete(self, proposal_id: int, issuer_address: str):
        """Delete proposal

        Args:
            proposal_id (int): proposal id
            issuer_address (str): delete the person's address

        Results:
            Delete the proposal's transaction

        """
        if issuer_address is None:
            issuer_address = self.tron.default_address.hex

        transaction = self.tron.transaction_builder.delete_proposal(
            proposal_id, issuer_address
        )
        sign = self.sign(transaction)
        response = self.broadcast(sign)

        return response

    def list_exchanges_paginated(self, limit=10, offset=0):
        """Paged query transaction pair list

        Args:
            limit (int): number of trading pairs  expected to be returned.
            offset (int): index of the starting trading pair

        """
        return self.tron.manager.request(
            "/wallet/listexchangespaginated", {"limit": limit, "offset": offset}
        )

    def get_node_info(self):
        """Get info about thre node"""
        return self.tron.manager.request("/wallet/getnodeinfo")

    def get_token_list_name(self, token_id: str) -> any:
        """Query token list by name.

        Args:
            token_id (str): The name of the token
        """
        if not is_string(token_id):
            raise ValueError("Invalid token ID provided")

        return self.tron.manager.request(
            "/wallet/getassetissuelistbyname", {"value": self.tron.toHex(text=token_id)}
        )

    def get_token_by_id(self, token_id: str) -> any:
        """Query token by id.

        Args:
            token_id (str): The id of the token, it's a string
        """
        if not is_string(token_id):
            raise ValueError("Invalid token ID provided")

        return self.tron.manager.request(
            "/wallet/getassetissuebyid", {"value": token_id}
        )

    def get_energy_prices(self):
        """Get Energy Prices."""

        return self.tron.manager.request("/wallet/getenergyprices")

    def get_bandwidth_prices(self):
        """Get Energy Prices."""

        return self.tron.manager.request("/wallet/getbandwidthprices")



#!/usr/bin/env python
# --------------------------------------------------------------------
# Copyright (c) iEXBase. All rights reserved.
# Licensed under the MIT License.
# See License.txt in the project root for license information.
# --------------------------------------------------------------------

"""
    setup
    =====

    Tron: A Python API for interacting with Tron (TRX)

    :copyright: © 2018 by the iEXBase.
    :license: MIT License
"""



import os
import pathlib
import platform
from setuptools import (
    find_packages,
    setup,
)
import shutil

py_version = platform.python_version()

PACKAGE_VERSION = "3.1.19"

EXTRAS_REQUIRE = {
    "tester": ["coverage", "pep8", "pyflakes", "pylint", "pytest-cov"],
    "docs": [
        "mock",
        "sphinx-better-theme>=0.1.4",
        "click>=5.1",
        "configparser==3.5.0",
        "contextlib2>=0.5.4",
        "py-solc>=0.4.0",
        "pytest>=2.7.2",
        "sphinx",
        "sphinx_rtd_theme>=0.1.9",
        "toposort>=1.4",
        "urllib3",
        "tronapi",
        "wheel >= 0.31.0",
    ],
    "dev": [
        "bumpversion",
        "flaky>=3.3.0",
        "hypothesis>=3.31.2",
        "pytest>=3.5.0,<4",
        "pytest-mock==1.*",
        "pytest-pythonpath>=0.3",
        "pytest-watch==4.*",
        "pytest-xdist==1.*",
        "setuptools>=38.6.0",
        "tox>=1.8.0",
        "twine >= 1.11.0",
        "tqdm",
        "when-changed",
    ],
}

EXTRAS_REQUIRE["dev"] = (
    EXTRAS_REQUIRE["tester"] + EXTRAS_REQUIRE["docs"] + EXTRAS_REQUIRE["dev"]
)

install_requires = [
    "toolz>=0.9.0,<1.0.0;implementation_name=='pypy'",
    "cytoolz>=0.9.0,<1.0.0;implementation_name=='cpython'",
    "eth-abi>=2.0.0b6,<3.0.0",
    "eth-account==0.4.0",
    "eth-utils>=1.3.0,<2.0.0",
    "eth-hash[pycryptodome]>=0.2.0,<1.0.0",
    "trx-utils",
    "hexbytes>=0.1.0,<1.0.0",
    "requests>=2.16.0,<3.0.0",
    "base58",
    "ecdsa",
    "attrdict",
]

this_dir = os.path.dirname(__file__)
readme_filename = os.path.join(this_dir, "README.rst")

PACKAGE_LONG_DESCRIPTION = pathlib.Path(readme_filename).read_text()

try:
    shutil.rmtree("./build")
    shutil.rmtree("./dist")
except OSError as e:
    print(f"Error: {e.filename} - {e.strerror}.")

setup(
    name="tronapi-JCO",
    version=PACKAGE_VERSION,
    description="A Python API for interacting with Tron (Forked from tronapi)",
    long_description=PACKAGE_LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    keywords="tron tron-api tron-api-python iexbase",
    url="https://github.com/RastinS/tron-api-JCO",
    author="Rustin Soraki",
    author_email="rustin.souraki@gmail.com",
    license="MIT License",
    zip_safe=False,
    python_requires=">=3.6,<4",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=["examples"]),
    include_package_data=True,
    install_requires=install_requires,
    tests_require=EXTRAS_REQUIRE["tester"],
    extras_require=EXTRAS_REQUIRE,
)

# python setup.py sdist bdist_wheel
# twine check dist/*
# twine upload dist/*
